import unittest
import grpc
from concurrent import futures
from hm_pyhelper.gateway_grpc.client import GatewayClient

from hm_pyhelper.protos import local_pb2
from hm_pyhelper.protos import local_pb2_grpc


class TestData:
    server_port = 4468
    validator_address_decoded = "11yJXQPG9deHqvw2ac6VWtNP7gZj8X3t3Qb3Gqm9j729p4AsdaA"
    pubkey_encoded = b"\x01\xc3\x06\x7f\xb9\x19}\xd1n2\xe2M\xeb\xb5\x11\x7f" \
                     b"\xbc\x12\xebT\xb9\x84R\xc7\xca\xf8o\xdddx\xea~\xab"
    pubkey_decoded = "14RdqcZC2rbdTBwNaTsj5EVWYaM7BKGJ44ycq6wWJy9Hg7RKCii"
    region_enum = 0
    region_name = "US915"
    expected_summary = {
        'region': region_name,
        'key': pubkey_decoded
    }


class MockServicer(local_pb2_grpc.apiServicer):
    def height(self, request, context):
        return TestData.height_res

    def region(self, request, context):
        return local_pb2.region_res(region=0)

    def pubkey(self, request, context):
        return local_pb2.pubkey_res(address=TestData.pubkey_encoded)

    def config(self, request, context):
        result = local_pb2.config_res()
        for key in request.keys:
            if key in TestData.chain_vars.keys():
                result.values.append(TestData.chain_vars.get(key))
            else:
                result.values.append(local_pb2.config_value(name=key))
        return result


class TestGatewayGRPCClient(unittest.TestCase):

    # we can start the real service hear by installing dpkg. But AFAIK
    # our testing methods, real service exposes us to random failures
    def setUp(self):
        self.mock_server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        local_pb2_grpc.add_apiServicer_to_server(MockServicer(), self.mock_server)
        self.mock_server.add_insecure_port(f'[::]:{TestData.server_port}')
        self.mock_server.start()

    def tearDown(self):
        self.mock_server.stop(None)

    def test_get_pubkey(self):
        with GatewayClient(f'localhost:{TestData.server_port}') as client:
            self.assertEqual(client.get_pubkey(), TestData.pubkey_decoded)

    def test_get_region(self):
        with GatewayClient(f'localhost:{TestData.server_port}') as client:
            self.assertEqual(client.get_region_enum(), TestData.region_enum)
            self.assertEqual(client.get_region(), TestData.region_name)

    def test_get_summary(self):
        with GatewayClient(f'localhost:{TestData.server_port}') as client:
            # summary when helium_gateway is not installed
            test_summary_copy = TestData.expected_summary.copy()
            self.assertIn(client.get_summary(),
                          [TestData.expected_summary, test_summary_copy])

    def test_connection_failure(self):
        with self.assertRaises(grpc.RpcError):
            with GatewayClient('localhost:1234') as client:
                client.get_pubkey()



"""
This module provides a set of functions to extract information about
the Single Board Computer in use.
It considers Balena environment variables as primary source of truth.
It also uses the device tree to extract information about the SBC.
"""

import os
from enum import Enum, auto
from collections import namedtuple

SBCInfo = namedtuple('SBCInfo', ['vendor_id', 'vendor_name', 'model_name'])


class DeviceVendorID(Enum):
    """
    Enum for device vendors.
    """
    INVALID = auto()
    ROCK_PI = auto()
    RASPBERRY_PI = auto()


# Pulled from
# https://www.balena.io/docs/reference/base-images/devicetypes/
BALENA_ENV_RASPBERRY_PI_MODELS = [
    'raspberry-pi',
    'raspberry-pi2',
    'raspberrypi3',
    'raspberrypi3-64',
    'raspberrypi4-64',
    'nebra-hnt',
    'raspberrypicm4-ioboard',
    'raspberrypi0-2w-64'
]

BALENA_ENV_ROCKPI_MODELS = ['rockpi-4b-rk3399']

BALENA_MODELS = {
    DeviceVendorID.ROCK_PI: BALENA_ENV_ROCKPI_MODELS,
    DeviceVendorID.RASPBERRY_PI: BALENA_ENV_RASPBERRY_PI_MODELS
}

COMMERCIAL_FLEETS = [
    156,  # Bobcat PX30
    56,  # Controllino
    106,  # COTX
    53,  # Finestra
    31,  # Nebra Indoor 868MHz
    40,  # Nebra Indoor RockPi 868MHz
    119,  # Nebra Indoor 915MHz
    58,  # Nebra Indoor RockPi 915MHz
    62,  # Linxdot
    42,  # Linxdot RKCM3
    143,  # Midas
    145,  # Nebra indoor1
    147,  # Nebra indoor2
    148,  # Nebra outdoor1
    149,  # Nebra outdoor2
    52,  # Helium OG
    80,  # Nebra Outdoor 868MHz
    107,  # Nebra Outdoor 915MHz
    47,  # PantherX
    66,  # Pisces
    73,  # Pycom
    88,  # RAK
    114,  # RisingHF
    124,  # Sensecap
    90,  # Syncrobit
    126,  # Syncrobit RKCM3
    98,  # Nebra Indoor Testing
    158,  # Bobcat PX30 Testing
    127,  # Controllino Testing
    87,  # COTX Testing
    76,  # Finestra Testing
    132,  # Linxdot Testing
    84,  # Linxdot RKCM3 Testing
    144,  # Midas Testing
    128,  # Helium OG Testing
    41,  # PantherX Testing
    43,  # Pisces Testing
    116,  # Pycom Testing
    113,  # RAK Testing
    103,  # RisingHF Testing
    60,  # Nebra RockPi Testing
    137,  # Sensecap Testing
    57,  # Syncrobit Testing
    111,  # Syncrobit RKCM3 Testing
    2006816,  # Rob Testing
    2061340,  # Rob Testing
]

NON_COMMERCIAL_FLEETS = [
    67,  # dev-1
    54,  # dev-2
    74,  # dev-3
    25,  # devnet-01
    26,  # devnet-02
    150,  # devnet-03
    153,  # dev-sensorhub
    121,  # Bobcat PX30
    105,  # Controllino
    136,  # COTX
    99,  # Disputed
    118,  # Finestra
    109,  # Indoor 470
    39,  # Indoor 868
    134,  # Indoor 868 RockPi
    91,  # Indoor 915
    122,  # Indoor 915 RockPi
    46,  # Linxdot
    68,  # Linxdot RKCM3
    138,  # Midas
    3,  # Indoor1
    27,  # Indoor2
    28,  # Outdoor1
    29,  # Outdoor2
    104,  # OG
    79,  # Outdoor 470
    63,  # Outdoor 868
    133,  # Outdoor 915
    94,  # Outdoor 915 RockPi
    32,  # Panther X1
    71,  # Pisces
    83,  # Pycom
    78,  # RAK
    85,  # RisingHF
    101,  # Sensecap
    130,  # Syncrobit
    69,  # Syncrobit RKCM3
    123,  # Testnet
    157,  # Bobcat PX30 Testnet
    102,  # Controllino Testnet
    86,  # COTX Testnet
    77,  # Finestra Testnet
    117,  # Linxdot Testnet
    110,  # Linxdot RKCM3 Testnet
    139,  # Midas Testnet
    70,  # OG Testnet
    129,  # Panther X1 Testnet
    125,  # Pisces Testnet
    131,  # Pycom Testnet
    36,  # RAK Testnet
    55,  # Rising HF Testnet
    95,  # RockPi Testnet
    75,  # Sensecap Testnet
    97,  # Syncrobit Testnet
    37,  # Syncrobit RKCM3 Testnet
    22,  # Indoor2 Testnet
]

NEBRA_API_URL = "https://api.cloud.nebra.com"

def device_model():
    with open('/proc/device-tree/model', 'r') as f:
        return f.readline().strip()


def sbc_info() -> SBCInfo:
    '''
    return SBCInfo formed by reading '/proc/device-tree/model'
    '''
    sbc_info = SBCInfo(vendor_id=DeviceVendorID.INVALID, vendor_name='', model_name='')
    dev_model = device_model()
    if dev_model.lower().find('raspberry') >= 0:
        sbc_info = SBCInfo(vendor_id=DeviceVendorID.RASPBERRY_PI,
                           vendor_name='Raspberry Pi',
                           model_name=dev_model)
    elif dev_model.lower().find('rock') >= 0:
        sbc_info = SBCInfo(vendor_id=DeviceVendorID.ROCK_PI,
                           vendor_name='Radxa Rock Pi',
                           model_name=dev_model)
    return sbc_info


def is_sbc_type(device_id: DeviceVendorID) -> bool:
    '''
    Return true if the sbc matches the type supplied.
    '''
    device_type = os.getenv('BALENA_DEVICE_TYPE')

    # use device tree supplied model name if evn not set
    if not device_type:
        return sbc_info().vendor_id == device_id

    # honor env override
    return device_type in BALENA_MODELS.get(device_id, [])


def is_commercial_fleet() -> bool:
    '''
    Return true if the device is in a commercial fleet. Otherwise return false.
    '''
    fleet_name = os.environ.get('BALENA_APP_NAME')
    fleet_id = int(os.environ.get('BALENA_APP_ID'))

    if not fleet_name.endswith('-c') or fleet_id not in COMMERCIAL_FLEETS:
        return False

    return True


def is_nebra_fleet() -> bool:
    '''
    Return true if the device is in a Nebra Cloud fleet. Otherwise return false.
    '''
    api_url = os.environ.get('BALENA_API_URL')
    fleet_id = int(os.environ.get('BALENA_APP_ID'))

    if api_url is not NEBRA_API_URL or (fleet_id not in COMMERCIAL_FLEETS and fleet_id not in NON_COMMERCIAL_FLEETS):
        return False

    return True 



# flake8: noqa
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: blockchain_txn_add_gateway_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#blockchain_txn_add_gateway_v1.proto\x12\x06helium\"\xbd\x01\n\x1d\x62lockchain_txn_add_gateway_v1\x12\r\n\x05owner\x18\x01 \x01(\x0c\x12\x0f\n\x07gateway\x18\x02 \x01(\x0c\x12\x17\n\x0fowner_signature\x18\x03 \x01(\x0c\x12\x19\n\x11gateway_signature\x18\x04 \x01(\x0c\x12\r\n\x05payer\x18\x05 \x01(\x0c\x12\x17\n\x0fpayer_signature\x18\x06 \x01(\x0c\x12\x13\n\x0bstaking_fee\x18\x07 \x01(\x04\x12\x0b\n\x03\x66\x65\x65\x18\x08 \x01(\x04\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'blockchain_txn_add_gateway_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _BLOCKCHAIN_TXN_ADD_GATEWAY_V1._serialized_start=48
  _BLOCKCHAIN_TXN_ADD_GATEWAY_V1._serialized_end=237
# @@protoc_insertion_point(module_scope)
